#
# @file icarus.tcl
# @brief Tcl Product commands and main flow
#
# Copyright by CAD Science, Inc
# All Rights Reserved.
# Written by Steve Cho
#
# Usage: tclsh icarus.tcl commandScript
#


#---------------------------------------------------------------------
# use the following instead of info exists
proc info_exists { name } {
	upvar $name lname
	if { ![info exists lname] || $lname == "" } {
		return 0
	} else {
		return 1
	}
}; # info_exists


#---------------------------------------------------------------------
# exit with message
# use this for errors before log file is assigned
proc Error0 { msg } {
	puts "Error: $msg\n"
	exit
}; # Error0


proc Error { msg } {
	global Product
	$Product Print "Error: $msg\n"
	exit
}; # Error


#---------------------------------------------------------------------
# check name extension to see if .gz
proc IsGz { name } {
	set ext [file extension $name]
	if { [string compare $ext ".gz"] } {
		return 0
	} else {
		return 1
	}
}


#---------------------------------------------------------------------
# set and return full path name with error checking
# return "" if error in GUI
# create directory if not there for writing
# curDir: full path
# write flag:
#  -1 - read without checking or message
#   0 - read
#   1 - write (error in console)
#   2 - write (error in log file)
proc SetFileName { curDir name write } {
	if { [file pathtype $name] == "relative" } {
		set name [file join $curDir $name]
	}
	if { $write > 0 } {
		set dName [file dirname $name]
		if { ![file isdirectory $dName] } {
			file mkdir $dName
		}
		# somehow this doesn't work in windows-xp
		#if { ![file exists $name] && ![file writable $name] } {
		#	if { $write > 1 } {
		#		Error "No write permission on $name !"
		#	} else {
		#		Error0 "No write permission on $name !"
		#	}
		#	return ""
		#}
	} elseif { $write == 0 } {
		if { ![file exists $name] } {
			Error "Input file $name doesn't exist !"
			return ""
		}
	}
	return $name
}; # SetFileName


#---------------------------------------------------------------------
# log message flags
proc SetReportDegree { deg } {
	global Product icReportDegree syntaxCheck
	if {!$syntaxCheck} {
		$Product Print "\nSetReportDegree $deg\n"
	}
	set icReportDegree $deg
	$Product\_tech configure -reportDegree $deg
}

proc GetReportDegree {} {
	global Product
	return [$Product\_tech cget -reportDegree]
}

proc SetReportCount { cnt } {
	global Product icReportCount syntaxCheck
	if {!$syntaxCheck} {
		$Product Print "SetReportCount $cnt\n"
	}
	set icReportCount $cnt
	$Product\_tech configure -reportCount $cnt
}

proc GetReport {} {
	global Product
	return [$Product\_tech cget -reportCount]
}


#---------------------------------------------------------------------
# layer set instead of TempLayer
proc SetLayer { name ld } {
	global syntaxCheck Product $name
	set iter [$Product cget -areaIter]
	if {!$syntaxCheck && $iter < 2} {
		$Product Print "\nSeLayer $name {$ld}\n"
	}
	set $name $ld
}


#---------------------------------------------------------------------
# for area
proc SetProcessSize { amount } {
	global Product syntaxCheck
	global icProcessSize
	set iter [$Product cget -areaIter]
	if {!$syntaxCheck && $iter < 2} {
		$Product Print "\nSetProcessSize $amount\n"
		set icProcessSize $amount
	}
}

proc SetProcessSizeX { amount } {
	global Product syntaxCheck
	global icProcessSizeX
	set iter [$Product cget -areaIter]
	if {!$syntaxCheck && $iter < 2} {
		$Product Print "\nSetProcessSizeX $amount\n"
		set icProcessSizeX $amount
	}
}

proc SetProcessSizeY { amount } {
	global Product syntaxCheck
	global icProcessSizeY
	set iter [$Product cget -areaIter]
	if {!$syntaxCheck && $iter < 2} {
		$Product Print "\nSetProcessSizeY $amount\n"
		set icProcessSizeY $amount
	}
}

proc SetAreaMargin { amount } {
	global Product syntaxCheck icAreaMarginM
	set iter [$Product cget -areaIter]
	if {!$syntaxCheck && $iter < 2} {
		$Product Print "\nSetAreaMargin $amount\n"
		set icAreaMarginM $amount
	}
	$Product\_tech configure -areaMargin $amount
}

proc GetAreaMargin {} {
	global Product
	return [$Product\_tech cget -areaMargin]
}

proc SetAreaMargin2 { amount } {
	global Product syntaxCheck
	set iter [$Product cget -areaIter]
	if {!$syntaxCheck && $iter < 2} {
		$Product Print "\nSetAreaMargin2 $amount\n"
	}
	$Product\_tech configure -areaMargin2 $amount
}


#---------------------------------------------------------------------
# Hierarchy Handling parameters
proc SetFlattenSize { amount } {
	global Product syntaxCheck
	if {!$syntaxCheck} {
		$Product Print "\nSetFlattenSize $amount\n"
	}
	$Product\_tech configure -flattenSize $amount
}

proc GetFlattenSize {} {
	global Product
	return [$Product\_tech cget -flattenSize]
}


#---------------------------------------------------------------------
# Log
#---------------------------------------------------------------------
proc Log { msg } {
	global syntaxCheck Product
	if {!$syntaxCheck} {
		$Product Print "$msg"
	}
}

# output message only for the first time in iteration or outside loop
proc Log1 { msg } {
	global syntaxCheck Product icSilent
	if {!$syntaxCheck && !$icSilent} {
		set iter [$Product cget -areaIter]
		if {$iter <= 1} {
			$Product Print "$msg"
		}
	}
}; # Log1


proc GetPath {} {
	set dir [pwd]
	return $dir
}


# Syntax: LogFile fileName
# set log file
proc LogFile { name } {
	global syntaxCheck
	if {$syntaxCheck} {
		return
	}

	global Product icLogFile
	if { [file pathtype $name] == "relative" } {
		set name [file join [GetPath] $name]
	}
	set icLogFile $name
	$Product Print "LogFile $name\n"
	if { ![$Product OpenLog $name] } {
		Error0 "LogFile"
	}
}; # LogFile


#---------------------------------------------------------------------
# Library
#---------------------------------------------------------------------
# Syntax: Library[Z] directoryName
#	Z for gzipped library
# define library directory
proc Library0 { name gz } {
	global syntaxCheck
	if {$syntaxCheck} {
		return
	}

	global Product CopyRight DLL Version BuildDate
	$Product Print "** Product   :\t$Product\n"
	$Product Print "** CopyRight :\t$CopyRight\n"
	$Product Print "** Executable:\t$DLL\n"
	$Product Print "** Version   :\t$Version\n"
	$Product Print "** Build Date:\t$BuildDate\n"

	if { ![file isdirectory $name] } {
		file mkdir $name
	}
	if { [file pathtype $name] == "relative" } {
		set name [file join [GetPath] $name]
	}
	if {$gz} {
		$Product Print "\nLibraryZ $name\n"
	} else {
		$Product Print "\nLibrary $name\n"
	}
	global icLib
	set icLib $name
	if { ![$Product OpenLibrary $name $gz] } {
		Error "Library"
	}

	# set tempdb
	set tname [file join $name temp]
	$Product SetTempDB $tname
}; # Library0


proc Library { name } {
	Library0 $name 0
}

proc LibraryZ { name } {
	Library0 $name 1
}; # Library


proc SetLibrary { id } {
	global syntaxCheck
	if {$syntaxCheck} {
		return
	}

	global Product
	Log1 "SetLibrary $id\n"
	$Product SetLibrary [expr $id-1]
}; # SetLibrary


#---------------------------------------------------------------------
# Syntax: LoadInfo
# load library's necessary information only, no geometry loading
proc LoadInfo {} {
	global syntaxCheck
	if {$syntaxCheck} {
		return
	}

	global Product
	$Product Print "LoadInfo\n"
	if { ![$Product LoadLibraryInfo] } {
		Error "LoadInfo"
	}
}; # LoadInfo


#---------------------------------------------------------------------
# Syntax: LoadAll
# load whole library data into memory
proc LoadAll {} {
	global syntaxCheck
	if {$syntaxCheck} {
		return
	}

	global Product
	$Product Print "LoadAll\n"
	if { ![$Product LoadLibraryAll 1] } {
		Error "LoadAll"
	}
}; # LoadAll


#---------------------------------------------------------------------
# GDS
#---------------------------------------------------------------------
# Syntax: GDSIn fileName[.gz]
#	.gz extension for gzip
# read GDS file into library
proc GDSIn { name args } {
	global syntaxCheck
	if {$syntaxCheck} {
		return
	}
	# avoid file checking for non-1st read
	SetFileName [pwd] $name 0

	global Product
	$Product Print "GDSIn $name\n"
	set gz [IsGz $name]
	#if { [file pathtype $name] == "relative" } {
	#	set name [file join [pwd] $name]
	#}
	# args is for optional prefix
	if { ![$Product LoadGds $name $gz $args] } {
		Error "GDSIn"
	}
}; # GDSIn


#---------------------------------------------------------------------
proc OASISIn { name args } {
	global syntaxCheck
	if {$syntaxCheck} {
		return
	}
	SetFileName [pwd] $name 0

	global Product
	$Product Print "OASISIn $name\n"
	# args is for optional prefix
	if { ![$Product LoadOasis $name $args] } {
		Error "OASISIn"
	}
}; # OASISIn


#---------------------------------------------------------------------
# Text Polygon
#---------------------------------------------------------------------
# Syntax: PolyIn fileName
# read test polygon file into library
proc PolyIn { name cname layer } {
	global syntaxCheck
	if {$syntaxCheck} {
		return
	}
	SetFileName [pwd] $name 0

	global Product
	$Product Print "PolyIn $name $cname $layer\n"
	#if { [file pathtype $name] == "relative" } {
	#	set name [file join [pwd] $name]
	#}
	if { ![$Product LoadPoly $name $cname [lindex $layer 0] [lindex $layer 1]] } {
		Error "PolyIn"
	}
}; # PolyIn


#---------------------------------------------------------------------
# Partial DB
#---------------------------------------------------------------------
# read partial DB file into library
proc DBIn { name cname } {
	global syntaxCheck
	if {$syntaxCheck} {
		return
	}
	SetFileName [pwd] $name 0

	global Product
	$Product Print "DBIn $name $cname\n"
	$Product SetTempDB $name
	$Product CopyTempDB $cname
	$Product CloseTempDB 
}; # DBIn


proc AppendBinaries { f1 f2 } {
	global syntaxCheck
	if {$syntaxCheck} {
		return
	}
	global Product
	$Product AppendBinaries $f1 $f2
}


#---------------------------------------------------------------------
# Set operation areas and/or layers
#---------------------------------------------------------------------
# Syntax: Areas { { cellName x1 y1 x2 y2 } ... }
#	cellName - set "" to process root cell (first root if input is forest)
#	x1/y1 - left bottom x/y coordinate in nanometer
#	x2/y2 - right top x/y coordinate in nanometer
#	set invalid area to process whole cell
# set operation areas
proc Areas { areas } {
	Log "Areas {$areas}\n"
	global icAreas
	set icAreas $areas
}


#---------------------------------------------------------------------
# Syntax: Layers { { layer1 datatype1 } ... }
# set operation layers
proc Layers { layers } {
	global syntaxCheck
	if {$syntaxCheck} {
		foreach layer $layers {
			if {[llength $layer] != 2} {
				puts "Wrong Layer Syntax $layer"
				exit
			}
		}
		return
	}

	global Product
	$Product ResetOpLayers
	set iter [$Product cget -areaIter]
	if {$iter < 2} {
		$Product Print "\nLayers \{\n"
	}
	foreach layer $layers {
		if {$iter < 2} {
			$Product Print "  {$layer}\n"
		}
		$Product AddOpLayer [lindex $layer 0] [lindex $layer 1]
	}
	if {$iter < 2} {
		$Product Print "\}\n"
	}
}; # Layers

proc ResetLayers {} {
	global syntaxCheck
	if {$syntaxCheck} {
		return
	}
	Log "ResetLayers\n"
	PlainOp ResetOpLayers
}

proc EnvLayers { layers } {
	global syntaxCheck
	if {$syntaxCheck} {
		foreach layer $layers {
			if {[llength $layer] != 2} {
				puts "Wrong Layer Syntax $layer"
				exit
			}
		}
		return
	}

	global Product
	$Product ResetOpLayers2
	$Product Print "\nEnvLayers {\n"
	foreach layer $layers {
		$Product Print "  {$layer}\n"
		$Product AddOpLayer2 [lindex $layer 0] [lindex $layer 1]
	}
	$Product Print "}\n"
}

proc ResetEnvLayers {} {
	PlainOp ResetOpLayers2
}

proc LayerMap { layermap } {
	global syntaxCheck
	if {$syntaxCheck} {
		foreach layers $layermap {
			if {[llength $layers] != 4} {
				puts "Wrong LayerMap Syntax $layers"
				exit
			}
		}
		return
	}

	global Product
	$Product ResetLayerMap
	$Product Print "\nLayerMap {\n"
	foreach layers $layermap {
		$Product Print "  {$layers}\n"
		$Product AddLayerMap [lindex $layers 0] [lindex $layers 1] [lindex $layers 2] [lindex $layers 3]
	}
	$Product Print "}\n"
}

proc ResetLayerMap {} {
	global syntaxCheck
	if {$syntaxCheck} {
		return
	}
	Log "ResetLayerMap\n"
	PlainOp ResetLayerMap
}


#---------------------------------------------------------------------
# split process areas
proc SetArea { area } {
	global Product icAreas icProcessSize icProcessSizeX icProcessSizeY
	# set default size to 25um
	set defaultSize 25000
	if {[info_exists icProcessSize]} {
		set icProcessSizeX $icProcessSize
		set icProcessSizeY $icProcessSize
	}
	if {![info_exists icProcessSizeX]} {
		set icProcessSizeX $defaultSize
	}
	if {![info_exists icProcessSizeY]} {
		set icProcessSizeY $defaultSize
	}
	set ProcessSizeX $icProcessSizeX
	set ProcessSizeY $icProcessSizeY
	
	set cname [lindex $area 0]
	set x1 [lindex $area 1]
	set y1 [lindex $area 2]
	set x2 [lindex $area 3]
	set y2 [lindex $area 4]
	if {$cname!="" && $x1 < $x2 && $y1 < $y2} {
		set nx [expr int(ceil(double($x2-$x1) / $ProcessSizeX))]
		set ny [expr int(ceil(double($y2-$y1) / $ProcessSizeY))]
       		if {$nx == 1 && $ny == 1} {
			# valid and small enough area
			# to skip message on 2 or more ForEachArea
			lappend icAreas [list $cname $x1 $y1 $x2 $y2]
			return
		}
	}
	if { ![$Product SetArea $cname $x1 $y1 $x2 $y2] } {
		return
	}
	# now split area
	set cname [$Product GetWorkingCell]
	set x1 [$Product GetAreaX1]
	set x2 [$Product GetAreaX2]
	set y1 [$Product GetAreaY1]
	set y2 [$Product GetAreaY2]
	set nx [expr int(ceil(double($x2-$x1) / $ProcessSizeX))]
	set ny [expr int(ceil(double($y2-$y1) / $ProcessSizeY))]
	if {$nx > 1 || $ny > 1} {
		set dx [expr int(ceil(double($x2-$x1) / $nx))]
		set dy [expr int(ceil(double($y2-$y1) / $ny))]
		set minx $x1
		for {set i 0} {$i<$nx} {incr i} {
			set maxx [expr $minx+$dx]
			if {$x2 < $maxx} {
				set maxx $x2
			}
			set miny $y1
			for {set j 0} {$j<$ny} {incr j} {
				set maxy [expr $miny+$dy]
				if {$y2 < $maxy} {
					set maxy $y2
				}
				lappend icAreas [list $cname $minx $miny $maxx $maxy]
				set miny $maxy
			}
			set minx $maxx
		}
	} else {
		lappend icAreas [list $cname $x1 $y1 $x2 $y2]
	}
}; # SetArea

proc SetAreas { areas } {
	global icAreas 

	set icAreas {}
	foreach area $areas {
		SetArea $area
	}
}; # SetAreas


#---------------------------------------------------------------------
# Analyze and report hierarchy
proc HierarchyAnalysis {} {
	global Product syntaxCheck
	if {!$syntaxCheck} {
		$Product Print "HierarchyAnalysis\n"
		set margin [$Product\_tech cget -areaMargin]
		$Product\_tech configure -areaMargin -2
		HierarchyHandling
		$Product\_tech configure -areaMargin $margin
	}
}; # HierarchyAnalysis


#---------------------------------------------------------------------
# apply hierarchical handling
# inside comment format: {#...}
proc Hierarchical { cmds } {
	global Product syntaxCheck icAreaMarginM icAreaMargin
	if {!$syntaxCheck && ![info_exists icAreaMarginM]} {
		# set automatic area margin
		$Product Print "\n#SetAreaMargin $icAreaMargin\n"
		$Product\_tech configure -areaMargin $icAreaMargin
	}
	HierarchyHandling
	Hierarchical0 $cmds
	PlainOp RemoveEnvLayers
	$Product configure -hierarchical 0
}

proc Hierarchical0 { cmds } {
	global Product syntaxCheck

	if {$syntaxCheck} {
		foreach cmd $cmds {
			if { ![string match "#" [lindex $cmd 0]] } {
				eval $cmd
			}
		}
		return
	}
	set n [$Product TotalLeaf]
	$Product configure -hierarchical 1
	# to make area reports 100-200 times
	set k [expr int(ceil($n / double(200)))]
	$Product configure -areaReportIter $k
	$Product Print "\nTotal $n Processing Cells\n"
	set i 1
	set cname [$Product GetFirstLeaf]
	while {$cname>0} {
		$Product configure -areaIter $i
		if {$i==1 || [expr fmod($i,$k)]==0.0} {
			$Product Print "\nArea $i/$n:\n"
		}
		if { ![$Product SetArea  $cname 0 0 0 0] } {
			break
		}
		foreach cmd $cmds {
			if { ![string match "#" [lindex $cmd 0]] } {
				eval $cmd
			}
		}
		# remove all temporary layers
		ResetTempLayers
		if {[expr fmod($i,$k)]==0.0} {
			$Product ReportCount
		}
		set cname [$Product GetNextLeaf]
		incr i
	}
	$Product configure -areaIter 0

	# update for additional command set
	if {$syntaxCheck==0} {
		UpdateLibraries
	}
}; # Hierarchical


#---------------------------------------------------------------------
# apply flat handling by area
# inside comment format: {#...}
proc ForEachArea { cmds } {
	global syntaxCheck Product icAreas icNAreas icAreaMarginM icAreaMargin
	global icClear
	if { ![info_exists icClear] } {
		set icClear 5000
	}
	if { ![info_exists icAreas] } {
		set icAreas {{"" 0 0 0 0}}
	}
	if {$syntaxCheck==0} {
		SetAreas $icAreas
		if {![info_exists icAreaMarginM]} {
			# set automatic area margin
			$Product Print "\n#SetAreaMargin $icAreaMargin\n"
			$Product\_tech configure -areaMargin $icAreaMargin
		}
	}
	set n [llength $icAreas]
	# to make area reports 100-200 times
	set k [expr int(ceil($n / double(200)))]
	#set k10 [expr $k * 10]
	$Product configure -areaReportIter $k
	if {$syntaxCheck==0} {
		$Product Print "\nTotal $n Areas\n"
		$Product OpenTempDB
	}

	set pname [lindex [lindex $icAreas 0] 0]
	set i 1
	set act 1; # actual repeat number
	foreach area $icAreas {
		$Product configure -areaIter $i
		set cname [lindex $area 0]
		set x1 [lindex $area 1]
		set y1 [lindex $area 2]
		set x2 [lindex $area 3]
		set y2 [lindex $area 4]
		if {$syntaxCheck==0} {
			if {$i==1 || [expr fmod($i,$k)]==0.0} {
				$Product Print "\nArea $i/$n:\n"
			}
			if { $pname != $cname } {
				# DP doesn't work w/ different working cell
				$Product CloseTempDB
				$Product CopyTempDB $pname
				$Product OpenTempDB
			}
			set pname $cname
			if { ![$Product SetArea  $cname $x1 $y1 $x2 $y2] } {
				break
			}
			if {[expr fmod($act,$icClear)]==0.0} {
				$Product ClearMemory
			}
		}
		foreach cmd $cmds {
			if { ![string match "#" [lindex $cmd 0]] } {
				eval $cmd
			}
		}
		if {$syntaxCheck==0} {
			# remove all temporary layers
			ResetTempLayers
			if {[expr fmod($i,$k)]==0.0} {
				$Product ReportCount
			}
		}
		incr i
		incr act
	}
	$Product configure -areaIter 0
	
	# update for additional command set
	if {$syntaxCheck==0} {
		$Product CloseTempDB
		$Product CopyTempDB $pname
		# else add tempDB manually later
		UpdateLibraries
	}
}; # ForEachArea


#---------------------------------------------------------------------
# apply cell based ops
# inside comment format: {#...}
# Not done yet
proc ForEachCell { cmds } {
	global syntaxCheck Product
	if {$syntaxCheck==0} {
		set n [$Product TotalCell]
		$Product Print "\nTotal $n Cells\n"
		set k [expr int(ceil($n / double(1000)))]
	} else {
		set n 1
	}
	for {set i 0} {$i<$n} {incr i} {
		if {$syntaxCheck==0} {
			set cname [$Product GetCellName $i]
			set j [expr $i+1]
			if {$j==0 || [expr fmod($j,$k)]==0.0} {
				$Product Print "\nCell $j/$n:\n"
			}
			if { $cname == "" } {
				continue
			}
			if { ![$Product SetArea  $cname 0 0 0 0] } {
				break
			}
		}
		foreach cmd $cmds {
			if { ![string match "#" [lindex $cmd 0]] } {
				eval $cmd
			}
		}
		if {$syntaxCheck==0} {
			# remove all temporary layers
			ResetTempLayers
		}
	}
	
	# update for additional command set
	if {$syntaxCheck==0} {
		UpdateLibraries
	}
}; # ForEachCell


#---------------------------------------------------------------------
# Template for commands

#---------------------------------------------------------------------
# double layer operation
# returns new layer datatype of temporary storage
proc DoubleOp { op ld1 ld2 } {
	if {[llength $ld1] == 1} { # non-list
		upvar #0 $ld1 ldt1
	} else {
		set l1 [lindex $ld1 0]
		if {[string is integer $l1]} {
			set ldt1 $ld1
		} else { # execute
			set ldt1 [eval $ld1]
		}
	}
	if {[llength $ld2] == 1} { # non-list
		upvar #0 $ld2 ldt2
	} else {
		set l2 [lindex $ld2 0]
		if {[string is integer $l2]} {
			set ldt2 $ld2
		} else {
			set ldt2 [eval $ld2]
		}
	}
	global syntaxCheck
	if {$syntaxCheck} {
		return {0 0}
	}
	
	global Product
	return [list [$Product $op [lindex $ldt1 0] [lindex $ldt1 1] \
		[lindex $ldt2 0] [lindex $ldt2 1]] 0]
}; # DoubleOp


#---------------------------------------------------------------------
# single layer operation
proc SingleOp { op ld1 } {
	if {[llength $ld1] == 1} { # non-list
		upvar #0 $ld1 ldt1
	} else {
		set l1 [lindex $ld1 0]
		if {[string is integer $l1]} {
			set ldt1 $ld1
		} else { # execute
			set ldt1 [eval $ld1]
		}
	}
	global syntaxCheck
	if {$syntaxCheck} {
		return {0 0}
	}

	global Product
	return [list [$Product $op [lindex $ldt1 0] [lindex $ldt1 1]] 0]
}; # SingleOp


#---------------------------------------------------------------------
# no argument operation
proc PlainOp { op } {
	global syntaxCheck
	if {$syntaxCheck} {
		return
	}

	global Product
	return [$Product $op]
}


#---------------------------------------------------------------------
# Relational DB
# returns new layer datatype of temporary storage
proc RELATE { ld1 ld2 } {
	Log1 "RELATE $ld1 $ld2\n"
	return [DoubleOp Relate $ld1 $ld2]
}

proc ROR1 { ld1 } {
	Log1 "ROR1 $ld1\n"
	return [SingleOp ROR1 $ld1]
}

proc ROR2 { ld1 } {
	Log1 "ROR2 $ld1\n"
	return [SingleOp ROR2 $ld1]
}

proc ROR { ld1 } {
	Log1 "ROR $ld1\n"
	return [SingleOp ROR $ld1]
}

proc RAND1 { ld1 } {
	Log1 "RAND1 $ld1\n"
	return [SingleOp RAND1 $ld1]
}

proc RAND2 { ld1 } {
	Log1 "RAND2 $ld1\n"
	return [SingleOp RAND2 $ld1]
}

proc RAND { ld1 } {
	Log1 "RAND $ld1\n"
	return [SingleOp RAND $ld1]
}

proc RDIFF { ld1 } {
	Log1 "RDIFF $ld1\n"
	return [SingleOp RDIFF $ld1]
}

proc RDIFF2 { ld1 } {
	Log1 "RDIFF2 $ld1\n"
	return [SingleOp RDIFF2 $ld1]
}

proc RXOR { ld1 } {
	Log1 "RXOR $ld1\n"
	return [SingleOp RXOR $ld1]
}

proc SIN { ld1 } {
	Log1 "SIN $ld1\n"
	return [SingleOp SIN $ld1]
}

proc SIN2 { ld1 } {
	Log1 "SIN2 $ld1\n"
	return [SingleOp SIN2 $ld1]
}

proc SOUT { ld1 } {
	Log1 "SOUT $ld1\n"
	return [SingleOp SOUT $ld1]
}

proc SOUT2 { ld1 } {
	Log1 "SOUT2 $ld1\n"
	return [SingleOp SOUT2 $ld1]
}

proc TIN { ld1 } {
	Log1 "TIN $ld1\n"
	return [SingleOp TIN $ld1]
}

proc TIN2 { ld1 } {
	Log1 "TIN2 $ld1\n"
	return [SingleOp TIN2 $ld1]
}

proc TOUT { ld1 } {
	Log1 "TOUT $ld1\n"
	return [SingleOp TOUT $ld1]
}

proc TOUT2 { ld1 } {
	Log1 "TOUT2 $ld1\n"
	return [SingleOp TOUT2 $ld1]
}

proc IN { ld1 } {
	Log1 "IN $ld1\n"
	return [SingleOp INSIDE $ld1]
}

proc IN2 { ld1 } {
	Log1 "IN2 $ld1\n"
	return [SingleOp INSIDE2 $ld1]
}

proc OUT { ld1 } {
	Log1 "OUT $ld1\n"
	return [SingleOp OUTSIDE $ld1]
}

proc OUT2 { ld1 } {
	Log1 "OUT2 $ld1\n"
	return [SingleOp OUTSIDE2 $ld1]
}

proc OVER { ld1 } {
	Log1 "OVER $ld1\n"
	return [SingleOp OVER $ld1]
}

proc OVER2 { ld1 } {
	Log1 "OVER2 $ld1\n"
	return [SingleOp OVER2 $ld1]
}

proc Append { ld1 ld2 } {
	Log1 "Append $ld1 $ld2\n"
	return [DoubleOp Append $ld1 $ld2]
}

proc NOT { op ld1 } {
	# must run ld1 first and then run op
	if {[llength $ld1] == 1} { # non-list
		upvar #0 $ld1 ldt1
	} else {
		set l1 [lindex $ld1 0]
		if {[string is integer $l1]} {
			set ldt1 $ld1
		} else { # execute
			set ldt1 [eval $ld1]
		}
	}
	global syntaxCheck
	if {$syntaxCheck} {
		return {0 0}
	}
	
	global Product 
	Log1 "NOT $op $ld1\n"
	$Product configure -_not 1
	set ret [$op $ldt1]
	$Product configure -_not 0; # reset
	return $ret
}; # NOT


#---------------------------------------------------------------------
# BOOLEAN
# returns new layer datatype of temporary storage
#---------------------------------------------------------------------
proc MergeLayers {} {
	Log1 "MergeLayers\n"
	return [PlainOp MergeLayers]
}

# Syntax: OR1 {layer1 datatype1}
# merge layer
proc OR1 { ld1 } {
	Log1 "OR1 $ld1\n"
	return [SingleOp OR $ld1]
}

# Syntax: OR {layer1 datatype1} {layer2 datatype2}
proc OR { ld1 ld2 } {
	Log1 "OR $ld1 $ld2\n"
	return [DoubleOp OR $ld1 $ld2]
}

# Syntax: AND1 {layer1 datatype1}
# merge layer
proc AND1 { ld1 } {
	Log1 "AND1 $ld1\n"
	return [SingleOp AND $ld1]
}

# Syntax: AND layer1 datatype1 layer2 datatype2
proc AND { ld1 ld2 } {
	Log1 "AND $ld1 $ld2\n"
	return [DoubleOp AND $ld1 $ld2]
}

# Syntax: XOR layer1 datatype1 layer2 datatype2
proc XOR { ld1 ld2 } {
	Log1 "XOR $ld1 $ld2\n"
	return [DoubleOp XOR $ld1 $ld2]
}

# Syntax: DIFF layer1 datatype1 layer2 datatype2
# L1 - L2
proc DIFF { ld1 ld2 } {
	Log1 "DIFF $ld1 $ld2\n"
	return [DoubleOp DIFF $ld1 $ld2]
}


#---------------------------------------------------------------------
# double layer operation with one parameter
proc DoubleOpParam { op ld1 ld2 amount } {
	if {[llength $ld1] == 1} { # non-list
		upvar #0 $ld1 ldt1
	} else {
		set l1 [lindex $ld1 0]
		if {[string is integer $l1]} {
			set ldt1 $ld1
		} else { # execute
			set ldt1 [eval $ld1]
		}
	}
	if {[llength $ld2] == 1} { # non-list
		upvar #0 $ld2 ldt2
	} else {
		set l2 [lindex $ld2 0]
		if {[string is integer $l2]} {
			set ldt2 $ld2
		} else {
			set ldt2 [eval $ld2]
		}
	}
	global Product icAreaMargin
	if {$icAreaMargin < $amount} {
		set icAreaMargin $amount
	}
	global syntaxCheck
	if {$syntaxCheck} {
		return {0 0}
	}
	
	return [list [$Product $op [lindex $ldt1 0] [lindex $ldt1 1] \
		[lindex $ldt2 0] [lindex $ldt2 1] $amount] 0]
}; # DoubleOpParam

# with option
proc DoubleOpParam2 { op ld1 ld2 amount opt } {
	if {[llength $ld1] == 1} { # non-list
		upvar #0 $ld1 ldt1
	} else {
		set l1 [lindex $ld1 0]
		if {[string is integer $l1]} {
			set ldt1 $ld1
		} else { # execute
			set ldt1 [eval $ld1]
		}
	}
	if {[llength $ld2] == 1} { # non-list
		upvar #0 $ld2 ldt2
	} else {
		set l2 [lindex $ld2 0]
		if {[string is integer $l2]} {
			set ldt2 $ld2
		} else {
			set ldt2 [eval $ld2]
		}
	}
	global Product icAreaMargin
	if {$icAreaMargin < $amount} {
		set icAreaMargin $amount
	}
	global syntaxCheck
	if {$syntaxCheck} {
		return {0 0}
	}
	
	return [list [$Product $op [lindex $ldt1 0] [lindex $ldt1 1] \
		[lindex $ldt2 0] [lindex $ldt2 1] $amount $opt] 0]
}; # DoubleOpParam2


#---------------------------------------------------------------------
# single layer operation with one parameter
proc SingleOpParam { op ld1 amount } {
	if {[llength $ld1] == 1} { # non-list
		upvar #0 $ld1 ldt1
	} else {
		set l1 [lindex $ld1 0]
		if {[string is integer $l1]} {
			set ldt1 $ld1
		} else { # execute
			set ldt1 [eval $ld1]
		}
	}
	global Product icAreaMargin
	if { $amount < 0 } {
		set t [expr -$amount]
	} else {
		set t $amount
	} 
	if {$icAreaMargin < $t} {
		set icAreaMargin $t
	}
	global syntaxCheck
	if {$syntaxCheck} {
		return {0 0}
	}
	
	# produce strong message, works only in wish
	#set layer [lindex $ldt1 0]
	#set dt [lindex $ldt1 1]
	#if { $layer < 256 } {
	#	if { ![$Product LayerExist $layer $dt] } {
	#		showMsg "No layer {$layer $dt} exist" "Warning"
	#		return [list [$Product GetTempLayer] 0]
	#	}
	#}
	return [list [$Product $op [lindex $ldt1 0] [lindex $ldt1 1] $amount] 0]
}; # SingleOpParam

# with option or 2 parameter
proc SingleOpParam2 { op ld1 amount opt } {
	if {[llength $ld1] == 1} { # non-list
		upvar #0 $ld1 ldt1
	} else {
		set l1 [lindex $ld1 0]
		if {[string is integer $l1]} {
			set ldt1 $ld1
		} else { # execute
			set ldt1 [eval $ld1]
		}
	}
	global Product icAreaMargin
	if {$icAreaMargin < $amount} {
		set icAreaMargin $amount
	}
	global syntaxCheck
	if {$syntaxCheck} {
		return {0 0}
	}
	
	return [list [$Product $op [lindex $ldt1 0] [lindex $ldt1 1] $amount $opt] 0]
}; # SingleOpParam2


#---------------------------------------------------------------------
# Dimensional DRC
proc SPACE { ld1 args } {
	Log1 "SPACE $ld1 $args\n"

	set n [llength $args]
	set opti 0
	if {$n == 0} {
		puts "Wrong parameters in SPACE command: $ld1"
		exit
	} elseif {$n == 1} {
		# must be amount
		return [SingleOpParam2 SPACE $ld1 [lindex $args 0] $opti]
	} elseif {$n == 2} {
		set opt [lindex $args 1]
		if {[string is integer $opt]} {
			# opt is amount
			# 2 layer spacing w/o option
			return [DoubleOpParam2 SPACE $ld1 [lindex $args 0] $opt $opti]
		} else {
			# 1 layer spacing w/ option
			if {[string equal $opt "corner"]} {
				set opti 1
			} elseif {[string equal $opt "edge"]} {
				set opti 2
			} elseif {[string equal $opt "parallel"]} {
				set opti 3
			}
			return [SingleOpParam2 SPACE $ld1 [lindex $args 0] $opti]
		}
	} else {
		# 2 layer spacing w/ option
		set opt [lindex $args 2]
		if {[string equal $opt "corner"]} {
			set opti 1
		} elseif {[string equal $opt "edge"]} {
			set opti 2
		} elseif {[string equal $opt "parallel"]} {
			set opti 3
		}
		return [DoubleOpParam2 SPACE $ld1 [lindex $args 0] [lindex $args 1] $opti]
	}
}; # SPACE

proc WIDTH { ld1 amount args } {
	set n [llength $args]
	set opti 0
	if {$n == 1} {
		# optional
		set opt [lindex $args 0]
		if {[string equal $opt "corner"]} {
			set opti 1
		} elseif {[string equal $opt "edge"]} {
			set opti 2
		} elseif {[string equal $opt "parallel"]} {
			set opti 3
		}
	}
	Log1 "WIDTH $ld1 $amount $args\n"
	return [SingleOpParam2 WIDTH $ld1 $amount $opti]
}

proc ENC { ld1 ld2 amount args } {
	set n [llength $args]
	set opti 0
	if {$n == 1} {
		# optional
		set opt [lindex $args 0]
		if {[string equal $opt "corner"]} {
			set opti 1
		} elseif {[string equal $opt "edge"]} {
			set opti 2
		} elseif {[string equal $opt "parallel"]} {
			set opti 3
		}
	}
	Log1 "ENC $ld1 $ld2 $amount $args\n"
	return [DoubleOpParam2 Enc $ld1 $ld2 $amount $opti]
}


#---------------------------------------------------------------------
# Syntax: Inverse layer
proc Inverse { ld1 } {
	Log1 "Inverse $ld1\n"
	return [SingleOp Inverse $ld1]
}

#---------------------------------------------------------------------
# Syntax: SIZING layer datatype amount
proc SIZING { ld1 amount } {
	Log1 "SIZING $ld1 $amount\n"
	return [SingleOpParam SIZING $ld1 $amount]
}


#---------------------------------------------------------------------
# set temporary layer to newLayer name 
# make datatype 1
proc TempLayer { newLayer layerdt } {
	global $newLayer Product icSilent
	Log1 "TempLayer $newLayer { $layerdt }\n"
	set l1 [lindex $layerdt 0]
	if {[string is integer $l1]} {
		set nlayer { [lindex $layerdt 0] 1 }
	} else { # execute command
		set old $icSilent
		set icSilent 1
		set nlayer [list [lindex [eval $layerdt] 0] 1]
		set icSilent $old
	}
	set $newLayer $nlayer
	global syntaxCheck icReportDegree
	if {!$syntaxCheck} {
		if {[info_exists icReportDegree] && $icReportDegree > 0} {
			$Product Print "TempLayer $newLayer {$nlayer}\n"
		}
	}

	return $nlayer
}

# this works only if layer is given in reverse order (from last to first)
proc ResetTempLayer { layer } {
	if {[llength $layer] == 1} { # non-list
		global $layer
		upvar #0 $layer ldt1
	} else {
		puts "Syntax Error: ResetTempLayer: $layer"
		exit
	}
	global syntaxCheck
	if {$syntaxCheck} {
		unset $layer
		return
	}

	global Product
	Log1 "ResetTempLayer $layer\n"
	$Product ResetTempLayer [lindex $ldt1 0]
	unset $layer
}

proc ResetTempLayers {} {
	Log1 "ResetTempLayers\n"
	PlainOp ResetTempLayers
}


#---------------------------------------------------------------------
proc SaveDB { ld1 layerdt2 } {
	Log1 "SaveDB { $ld1 } { $layerdt2 }\n"
	global icSilent
	set old $icSilent
	set icSilent 1
	if {[llength $ld1] == 1} { # non-list
		upvar #0 $ld1 ldt1
	} else {
		set l1 [lindex $ld1 0]
		if {[string is integer $l1]} {
			set ldt1 $ld1
		} else { # execute
			set ldt1 [eval $ld1]
		}
	}
	set icSilent $old
	set layer [lindex $ldt1 0]
	global Product
	if {$layer > 0 && $layer <= 255} {
		puts $layer
		$Product Print "Syntax Error: must use TempLayer for SaveDB\n"
		exit
	}
	global syntaxCheck
	if {$syntaxCheck} {
		return
	}

	$Product SaveDB [lindex $ldt1 0] [lindex $ldt1 1] [lindex $layerdt2 0] [lindex $layerdt2 1]
}; # SaveDB


#---------------------------------------------------------------------
proc SavePoly { ld1 name } {
	Log1 "SavePoly { $ld1 } $name\n"
	global icSilent
	set old $icSilent
	set icSilent 1
	if {[llength $ld1] == 1} { # non-list
		upvar #0 $ld1 ldt1
	} else {
		set l1 [lindex $ld1 0]
		if {[string is integer $l1]} {
			set ldt1 $ld1
		} else { # execute
			set ldt1 [eval $ld1]
		}
	}
	set icSilent $old
	set layer [lindex $ldt1 0]
	global Product
	if {$layer > 0 && $layer <= 255} {
		$Product Print "Syntax Error: must use TempLayer for SavePoly\n"
		exit
	}
	global syntaxCheck
	#if { ![file isdirectory $dir] } {
	#	file mkdir $dir
	#}
	#set iter [$Product cget -areaIter]
	set fname [SetFileName [GetPath] $name 2]
	if {$syntaxCheck} {
		file delete $fname
		return
	}
	$Product SavePoly $layer [lindex $ldt1 1] "$fname"
}; # SavePoly


#---------------------------------------------------------------------
proc HierarchyHandling {} {
	PlainOp HierarchyHandling
}

#---------------------------------------------------------------------
proc UpdateLibrary {} {
	PlainOp UpdateLibrary
}

proc UpdateLibraries {} {
	Log "\nUpdateLibraries\n"
	PlainOp UpdateLibraries
}


proc CloseLibrary {} {
	PlainOp CloseLibrary
}

proc CloseLibraries {} {
	PlainOp CloseLibraries
}



#---------------------------------------------------------------------
# Main
# check if input script is given
if { $argc == 0 } {
	puts "Input script must be provided !"
	exit
}

# set environments
#if { [info exists env(ICARUS)] } {
#	set productDir $env(ICARUS)
#} else {
#	puts "Environment variable ICARUS not set !"
#	exit
#}

# load commands tcl
set TclDir [file dirname $argv0]
if { $TclDir == "." } {
	set TclDir [pwd]
}

# load binary (dll/so)
set productDir $TclDir
if { $tcl_platform(platform) == "windows" } {
	set DLL [file join $productDir icarus.dll]
	load $DLL
} else {
	set DLL $env(ICARUS)
	set DLL [file join $productDir $DLL]	
}
API $Product
Tech $Product\_tech -this [$Product cget -tech]


# first, syntax check of user script
set syntaxCheck 1
set icSilent 0
set icAreaMargin 0
if { $argc == 1 } {
	source $argv
} else {
	source [lindex $argv 0]
}
if {[info_exists icAreas]} {
	unset icAreas
}
set syntaxCheck 0
puts "Passed Syntax Check"

# execute user script
if { $argc == 1 } {
	source $argv
} else {
	source [lindex $argv 0]
}


# close
$Product UpdateLibraries
$Product CloseLibraries
$Product CloseLog

